# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..client import Client
    from typing import Dict, List, Set, Union, Tuple


def construct_client_dict(client: Client, data: Dict) -> Dict:
    # TODO: fix docs
    """

    Parameters
    ----------
    client
    data

    Returns
    -------

    """
    return {**data, "_client": client}


def remove_none(obj: Union[List, Dict, Set, Tuple]) -> Union[List, Dict, Set, Tuple]:
    """
    Removes all ``None`` values from a list, dict or set.

    Parameters
    ----------
    obj : Union[List, Dict, Set, Tuple]
        The list, dict, set or tuple to remove ``None`` values from.

    Returns
    -------
    Union[List, Dict, Set, Tuple]
        The list, dict, set or tuple, without ``None`` values.
    """
    if isinstance(obj, list):
        return [i for i in obj if i is not None]
    elif isinstance(obj, tuple):
        return tuple(i for i in obj if i is not None)
    elif isinstance(obj, set):
        return obj - {None}
    elif isinstance(obj, dict):
        return {k: v for k, v in obj.items() if None not in (k, v)}



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from typing import Any, List, Tuple, Union, T

from ..utils.types import MISSING
from ..objects.app.command import AppCommandOptionChoice


class _CommandTypeMeta(type):
    def __getitem__(cls, args: Union[Tuple, Any]):
        if not isinstance(args, tuple):
            args = (args,)

        return cls(*args)


class CommandArg(metaclass=_CommandTypeMeta):
    """
    Holds the parameters of an application command option

    .. code-block:: python3

        CommandArg[
            # This is the type of command.
            # Supported types are str, int, bool, float, User, Channel, and Role
            int,
            # The modifiers to the command go here
            Description["Pick a number 1-10"],
            MinValue[1],
            MaxValue[10]
        ]

    Parameters
    ----------
    command_type : T
        The type of the command
    \\*args : :class:`~pincer.commands.arg_types.Modifier`

    """

    def __init__(self, command_type, *args):
        self.command_type = command_type
        self.modifiers = args

    def get_arg(self, arg_type: T) -> T:
        for arg in self.modifiers:
            if isinstance(arg, arg_type):
                return arg.get_payload()

        return MISSING


class Modifier(metaclass=_CommandTypeMeta):
    """
    Modifies a CommandArg by being added to
    :class:`~pincer.commands.arg_types.CommandArg`'s args.
    """


class Description(Modifier):
    """
    Represents the description of an application command option

    .. code-block:: python3

        # Creates an int argument with the description "example description"
        CommandArg[
            int,
            Description["example description"]
        ]

    Parameters
    ----------
    desc : str
        The description for the command.
    """

    def __init__(self, desc):
        self.desc = str(desc)

    def get_payload(self) -> str:
        return self.desc


class Choice(Modifier):
    """
    Represents a choice that the user can pick from

    .. code-block:: python3

        Choices[
            Choice["First Number", 10],
            Choice["Second Number", 20]
        ]

    Parameters
    ----------
    name : str
        The name of the choice
    value : Union[int, str, float]
        The value of the choice
    """

    def __init__(self, name, value):
        self.name = name
        self.value = value


class Choices(Modifier):
    """
    Represents a group of application command choices that a user can pick from

    .. code-block:: python3

        CommandArg[
            int,
            Choices[
                Choice["First Number", 10],
                20,
                50
            ]
        ]

    Parameters
    ----------
    \\*choices : Union[:class:`~pincer.commands.arg_types.Choice`, str, int, float]
        A choice. If the type is not :class:`~pincer.commands.arg_types.Choice`,
        the same value will be used for the choice name and value.
    """

    def __init__(self, *choices):
        self.choices = []

        for choice in choices:
            if isinstance(choice, Choice):
                self.choices.append(
                    AppCommandOptionChoice(name=choice.name, value=choice.value)
                )
                continue

            self.choices.append(
                AppCommandOptionChoice(name=str(choice), value=choice)
            )

    def get_payload(self) -> List[AppCommandOptionChoice]:
        return self.choices


class ChannelTypes(Modifier):
    """
    Represents a group of channel types that a user can pick from

    .. code-block:: python3

        CommandArg[
            Channel,
            # The user will only be able to choice between GUILD_TEXT and
            GUILD_TEXT channels.
            ChannelTypes[
                ChannelType.GUILD_TEXT,
                ChannelType.GUILD_VOICE
            ]
        ]

    Parameters
    ----------
    \\*types : :class:`~pincer.objects.guild.channel.ChannelType`
        A list of channel types that the user can pick from.
    """

    def __init__(self, *types):
        self.types = types

    def get_payload(self):
        return self.types


class MaxValue(Modifier):
    """
    Represents the max value for a number

    .. code-block:: python3

        CommandArg[
            int,
            # The user can't pick a number above 10
            MaxValue[10]
        ]

    Parameters
    ----------
    max_value : Union[:class:`float`, :class:`int`]
        The max value a user can choose.
    """

    def __init__(self, max_value):
        self.max_value = max_value

    def get_payload(self):
        return self.max_value


class MinValue(Modifier):
    """
    Represents the minimum value for a number

    .. code-block:: python3

        CommandArg[
            int,
            # The user can't pick a number below 10
            MinValue[10]
        ]

    Parameters
    ----------
    min_value : Union[:class:`float`, :class:`int`]
        The minimum value a user can choose.
    """

    def __init__(self, min_value):
        self.min_value = min_value

    def get_payload(self):
        return self.min_value



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum, IntEnum
from typing import TYPE_CHECKING, overload

from ...exceptions import EmbedOverflow
from ...utils.api_object import APIObject
from ...utils.conversion import construct_client_dict
from ...utils.types import MISSING

if TYPE_CHECKING:
    from typing import List, Optional

    from ..message.attachment import Attachment
    from ..message.component import MessageComponent
    from ..message.embed import Embed
    from ..message.user_message import UserMessage
    from ..message.user_message import AllowedMentions
    from ..user.user import User
    from ..guild.guild import Guild
    from ..guild.channel import Channel
    from ...utils.types import APINullable
    from ...utils.snowflake import Snowflake
    from ...client import Client


class WebhookCompatibility(Enum):
    GitHub = "github"
    Slack = "slack"
    Default = ""


class WebhookType(IntEnum):
    """Represents the type of webhook.

    Attributes
    ----------
    INCOMING:
        Incoming Webhooks can post messages to channel with a
        generated token.
    CHANNEL_FOLLOWER:
        Channel Follower Webhooks are internal webhooks used with
        Channel Following to post new messages into channels.
    APPLICATION:
        Application webhooks are webhooks used with Interactions
    """
    INCOMING = 1
    CHANNEL_FOLLOWER = 2
    APPLICATION = 3


@dataclass(repr=False)
class Webhook(APIObject):
    """Represents a Discord channel webhook.

    Attributes
    ----------
    id: :class:`~pincer.utils.snowflake.Snowflake`
        The id of the webhook
    type: :class:`~pincer.objects.guild.webhook.WebhookType`
        The type of the webhook
    channel_id: Optional[:class:`~pincer.utils.snowflake.Snowflake`]
        The channel id this webhook is for, if any
    name: Optional[:class:`str`]
        The default name of the webhook
    avatar: Optional[:class:`str`]
        The default user avatar hash of the webhook
    application_id: Optional[:class:`~pincer.utils.snowflake.Snowflake`]
        The bot/OAuth2 application that created this webhook
    user: APINullable[:class:`~pincer.objects.user.user.User`]
        The user this webhook was created by
        (not returned when getting a webhook with its token)
    token: APINullable[:class:`str`]
        The secure token of the webhook
        (returned for Incoming Webhooks)
    source_guild: APINullable[:class:`~pincer.objects.guild.guild.Guild`]
        The guild of the channel that this webhook is following
        (returned for Channel Follower Webhooks)
    source_channel: APINullable[:class:`~pincer.objects.guild.channel.Channel`]
        The channel that this webhook is following
        (returned for Channel Follower Webhooks)
    url: APINullable[:class:`str`]
        The url used for executing the webhook
        (returned by the webhooks OAuth2 flow)
    guild_id: APINullable[Optional[:class:`~pincer.objects.guild.guild.Guild`]]
        The guild id this webhook is for, if any
    """
    id: Snowflake
    type: WebhookType

    channel_id: Optional[Snowflake] = None
    name: Optional[str] = None
    avatar: Optional[str] = None
    application_id: Optional[Snowflake] = None

    user: APINullable[User] = MISSING
    token: APINullable[str] = MISSING
    source_guild: APINullable[Guild] = MISSING
    source_channel: APINullable[Channel] = MISSING
    url: APINullable[str] = MISSING

    guild_id: APINullable[Optional[Snowflake]] = MISSING

    async def edit(
        self,
        *,
        name: Optional[str] = None,
        avatar: Optional[str] = None,
        channel_id: Optional[Snowflake] = None,
        token: Optional[str] = None
    ) -> Webhook:
        """
        Modifies a webhook and returns it.
        Requires the ``MANAGE_WEBHOOKS`` permission.

        Parameters
        ----------
        name: Optional[:class:`str`]
            The new name of the webhook
        avatar: Optional[:class:`str`]
            The new avatar hash of the webhook
        channel_id: Optional[:class:`~pincer.utils.snowflake.Snowflake`]
            The new channel id this webhook is for
        token: Optional[:class:`str`]
            The new token of the webhook
        """
        request_route = (
            f"webhooks/{self.id}"
            + (f"/{token}" if token else "")
        )
        request_data = {
            "name": name,
            "avatar": avatar,
            "channel_id": channel_id
        }

        if token:
            del request_data["channel_id"]

        data = await self._http.patch(
            request_route,
            data=request_data
        )
        return Webhook.from_dict(
            construct_client_dict(self._client, data)
        )

    async def delete(self, token: Optional[str] = None):
        """
        Deletes a webhook.
        Requires the ``MANAGE_WEBHOOKS`` permission.

        Parameters
        ----------
        token: Optional[:class:`str`]
            The token of the webhook
        """
        await self._http.delete(
            f"webhooks/{self.id}"
            + (f"/{token}" if token else "")
        )

    @overload
    async def execute(
        self,
        webhook_compatibility: WebhookCompatibility = WebhookCompatibility.Default,  # noqa: E501
        *,
        thread_id: Optional[Snowflake] = None,
        wait: Optional[bool] = None,
        content: Optional[str] = None,
        username: Optional[str] = None,
        avatar_url: Optional[str] = None,
        tts: Optional[bool] = None,
        embeds: Optional[List[Embed]] = None,
        allowed_mentions: Optional[AllowedMentions] = None,
        components: Optional[List[MessageComponent]] = None,
        files: Optional[str] = None,  # TODO: Add support for files
        payload_json: Optional[str] = None,
        attachments: Optional[List[Attachment]] = None
    ):
        """|coro|
        Executes a webhook.

        Note that when sending a message, you must provide a value
        for at least one of ``content``, ``embeds``, or ``file``.

        Parameters
        ----------
        webhook_compatibility: :class:`~pincer.objects.guild.webhook.WebhookCompatibility`
            The compatibility of the webhook
        thread_id: Optional[:class:`~pincer.utils.snowflake.Snowflake`]
            ID of the thread to send message in
        wait: Optional[:class:`bool`]
            Waits for server confirmation of message send before
            response (defaults to ``true``, when ``false`` a message
            that is not saved does not return an error)
        content: Optional[:class:`str`]
            The message contents (up to 2000 characters)
        username: Optional[:class:`str`]
            Override the default username of the webhook
        avatar_url: Optional[:class:`str`]
            Override the default avatar of the webhook
        tts: Optional[:class:`bool`]
            True if this is a TTS message
        embeds: Optional[List[:class:`~pincer.objects.message.embed.Embed`]]
            Embedded ``rich`` content, up to 10 embeds
        allowed_mentions: Optional[:class:`~pincer.objects.message.user_message.AllowedMentions`]
            Allowed mentions for the message
        components: Optional[List[:class:`~pincer.objects.message.component.MessageComponent`]]
            The components to include in the message
        files: Optional[:class:`str`]
            The contents of the file being sent
        payload_json: Optional[:class:`str`]
            JSON encoded body of non-file params
        attachments: Optional[List[:class:`~pincer.objects.message.attachment.Attachment`]]
            Attachment objects with filename and description
        """
        ...

    async def execute(
        self,
        webhook_compatibility: WebhookCompatibility = WebhookCompatibility.Default,  # noqa: E501
        *,
        thread_id: Optional[Snowflake] = None,
        wait: Optional[bool] = None,
        **kwargs
    ):
        if len(kwargs.get("embeds", [])) > 10:
            raise EmbedOverflow("You can only include up to 10 embeds")

        request_route = f"webhooks/{self.id}/{self.token}"

        # Adding the subdirectory
        if webhook_compatibility.value:
            request_route += f"/{webhook_compatibility.value}"

        # Adding query params
        if wait is not None:
            request_route += f"?{wait=}"
        if thread_id is not None:
            request_route += "&?"[wait is None] + f"{thread_id=}"

        if webhook_compatibility == WebhookCompatibility.Default:
            request_data = kwargs
        else:
            request_data = None

        await self._http.post(request_route, data=request_data)

    async def execute_github(
        self,
        *,
        thread_id: Optional[Snowflake] = None,
        wait: Optional[bool] = None
    ):
        """|coro|
        Executes a GitHub compatible webhook.

        Parameters
        ----------
        thread_id: Optional[:class:`~pincer.utils.snowflake.Snowflake`]
            ID of the thread to send message in
        wait: Optional[:class:`bool`]
            Waits for server confirmation of message send before
            response (defaults to ``true``, when ``false`` a message
            that is not saved does not return an error)
        """
        await self.execute(
            WebhookCompatibility.GitHub,
            thread_id=thread_id,
            wait=wait
        )

    async def execute_slack(
        self,
        *,
        thread_id: Optional[Snowflake] = None,
        wait: Optional[bool] = None
    ):
        """|coro|
        Executes a Slack compatible webhook.

        Parameters
        ----------
        thread_id: Optional[:class:`~pincer.utils.snowflake.Snowflake`]
            ID of the thread to send message in
        wait: Optional[:class:`bool`]
            Waits for server confirmation of message send before
            response (defaults to ``true``, when ``false`` a message
            that is not saved does not return an error)
        """
        await self.execute(
            WebhookCompatibility.Slack,
            thread_id=thread_id,
            wait=wait
        )

    async def get_message(
        self,
        message_id: Snowflake,
        thread_id: Snowflake
    ) -> UserMessage:
        """|coro|
        Returns a previously-sent webhook message from the same token.

        Parameters
        ----------
        message_id: :class:`~pincer.utils.snowflake.Snowflake`
            The ID of the message to get
        thread_id: :class:`~pincer.utils.snowflake.Snowflake`
            The ID of the thread to get the message from

        Returns
        -------
        :class:`~pincer.objects.message.message.Message`
            The message
        """
        return UserMessage.from_dict(
            construct_client_dict(
                self._client,
                await self._http.get(
                    f"webhooks/{self.id}/{self.token}/messages/{message_id}",
                    params={"thread_id": thread_id}
                )
            )
        )

    async def delete_message(
        self,
        message_id: Snowflake,
        thread_id: Snowflake
    ):
        """|coro|
        Deletes a message created by a webhook.

        Parameters
        ----------
        message_id: :class:`~pincer.utils.snowflake.Snowflake`
            The ID of the message to delete
        thread_id: :class:`~pincer.utils.snowflake.Snowflake`
            The ID of the thread to delete the message from
        """
        await self._http.delete(
            f"webhooks/{self.id}/{self.token}/messages/{message_id}"
            + (f"?{thread_id=}" if thread_id else "")
        )

    @overload
    async def edit_message(
        self,
        message_id: Snowflake,
        *,
        thread_id: Optional[Snowflake] = None,
        content: Optional[str] = None,
        embeds: Optional[List[Embed]] = None,
        allowed_mentions: Optional[AllowedMentions] = None,
        components: Optional[List[MessageComponent]] = None,
        files: Optional[str] = None,  # TODO: Add support for files
        payload_json: Optional[str] = None,
        attachments: Optional[List[Attachment]] = None
    ) -> UserMessage:
        """|coro|
        Edits a previously-sent webhook message from the same token.

        Parameters
        ----------
        message_id: :class:`~pincer.utils.snowflake.Snowflake`
            The ID of the message to edit
        thread_id: Optional[:class:`~pincer.utils.snowflake.Snowflake`]
            ID of the thread the message is in
        content: Optional[:class:`str`]
            The new content of the message (up to 2000 characters)
        embeds: Optional[List[:class:`~pincer.objects.message.embed.Embed`]]
            Embedded ``rich`` content, up to 10 embeds
        allowed_mentions: Optional[:class:`~pincer.objects.message.user_message.AllowedMentions`]
            Allowed mentions for the message
        components: Optional[List[:class:`~pincer.objects.message.component.MessageComponent`]]
            The components to include in the message
        files: Optional[:class:`str`]
            The contents of the file being sent/edited
        payload_json: Optional[:class:`str`]
            JSON encoded body of non-file params
            (multipart/form-data only)
        attachments: Optional[List[:class:`~pincer.objects.message.attachment.Attachment`]]
            Attached files to keep and
            possible descriptions for new files
        """
        ...

    async def edit_message(
        self,
        message_id: Snowflake,
        *,
        thread_id: Optional[Snowflake] = None,
        **kwargs
    ) -> UserMessage:
        if len(kwargs.get("embeds", [])) > 10:
            raise EmbedOverflow("You can only include up to 10 embeds")

        data = await self._http.patch(
            f"webhooks/{self.id}/{self.token}/messages/{message_id}"
            + (f"?{thread_id=}" if thread_id else ""),
            data=kwargs
        )
        return UserMessage.from_dict(
            construct_client_dict(self._client, data)
        )

    @classmethod
    async def from_id(
        cls,
        client: Client,
        id: Snowflake,
        token: Optional[str] = None
    ) -> Webhook:
        """|coro|
        Gets a webhook by its ID.

        Parameters
        ----------
        client: `~pincer.client.Client`
            The client to use to make the request.
        id: `~pincer.utils.snowflake.Snowflake`
            The ID of the webhook to get.
        token: Optional[:class:`str`]
            The token of the webhook to get.

        Returns
        -------
        `~pincer.objects.guild.webhook.Webhook`
            The webhook with the given ID.
        """
        return cls.from_dict(
            construct_client_dict(
                client,
                await client.http.get(
                    f"webhooks/{id}"
                    + (f"/{token}" if token else "")
                )
            )
        )



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""
sent when the client receives a notification
(mention or new message in eligible channels)
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from ..objects.events.notification import NotificationCreateEvent
from ..utils.conversion import construct_client_dict
from ..utils.types import Coro

if TYPE_CHECKING:
    from ..client import Client
    from ..core.gateway import Gateway
    from ..core.gateway import GatewayDispatch


async def notification_create_middleware(
    self: Client,
    gateway: Gateway,
    payload: GatewayDispatch
):
    """|coro|

    Middleware for the ``on_notification_create`` event.

    Parameters
    ----------
    payload : :class:`~pincer.core.gateway.GatewayDispatch`
        The data received from the notification create event.
    gateway : :class:`~pincer.core.gateway.Gateway`
        The gateway for the current shard.

    Returns
    -------
    Tuple[:class:`str`, :class:`~pincer.objects.events.notification.NotificationCreateEvent`]
        ``on_notification_create`` and a ``NotificationCreateEvent``
    """  # noqa: E501
    channel_id = payload.data.get("channel_id")
    payload.data["message"]["channel_id"] = channel_id
    return (
        "on_notification_create",
        NotificationCreateEvent.from_dict(
            construct_client_dict(self, payload.data)
        ),
    )


def export() -> Coro:
    return notification_create_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""sent when a message is updated in a subscribed text channel"""

from __future__ import annotations

from typing import TYPE_CHECKING

from ..objects import UserMessage
from ..utils.conversion import construct_client_dict

if TYPE_CHECKING:
    from typing import Tuple

    from ..client import Client
    from ..core.gateway import Gateway
    from ..core.gateway import GatewayDispatch


async def message_update_middleware(
    self: Client, gateway: Gateway, payload: GatewayDispatch
) -> Tuple[str, UserMessage]:
    """|coro|


    Middleware for the ``on_message_update`` event.
        generate a class for the message that has been updated.

    Parameters
    ----------
    payload : :class:`~pincer.core.gateway.GatewayDispatch`
        The data received from the message update event event
    gateway : :class:`~pincer.core.gateway.Gateway`
        The gateway for the current shard.

    Returns
    -------
    Tuple[:class:`str`, :class:`~pincer.objects.message.user_message.UserMessage`]
        ``on_message_update`` and a ``UserMessage``
    """  # noqa: E501
    return (
        "on_message_update",
        UserMessage.from_dict(construct_client_dict(self, payload.data)),
    )


def export():
    return message_update_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""Sent when a channel is created/joined on the client."""
from __future__ import annotations

from typing import TYPE_CHECKING

from ..objects.guild.channel import Channel
from ..utils.conversion import construct_client_dict

if TYPE_CHECKING:
    from typing import Tuple
    from ..core.gateway import GatewayDispatch
    from ..client import Client
    from ..core.gateway import Gateway


async def channel_create_middleware(
    self: Client,
    gateway: Gateway,
    payload: GatewayDispatch
) -> Tuple[str, Channel]:
    """|coro|

    Middleware for the ``on_channel_creation`` event.

    Parameters
    ----------
    payload : :class:`~pincer.core.gateway.GatewayDispatch`
        The data received from the ready event.
    gateway : :class:`~pincer.core.gateway.Gateway`
        The gateway for the current shard.

    Returns
    -------
    Tuple[:class:`str`, List[:class:`~pincer.objects.guild.channel.Channel`]]
        ``on_channel_creation`` and a channel.
    """

    channel: Channel = Channel.from_dict(
        construct_client_dict(self, payload.data)
    )
    self.guilds[channel.guild_id].channels.append(channel)

    self.channels[channel.id] = channel

    return "on_channel_creation", channel


def export():
    return channel_create_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""
sent when a user's voice state changes in a subscribed voice channel
(mute, volume, etc.)
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from ..objects.events.webhook import WebhooksUpdateEvent
from ..utils.conversion import construct_client_dict

if TYPE_CHECKING:
    from ..client import Client
    from ..core.gateway import Gateway
    from ..core.gateway import GatewayDispatch


async def webhooks_update_middleware(
    self: Client,
    gateway: Gateway,
    payload: GatewayDispatch
):
    """|coro|

    Middleware for the ``on_webhooks_update`` event.

    Parameters
    ----------
    payload : :class:`~pincer.core.gateway.GatewayDispatch`
        The data received from the webhooks update event.
    gateway : :class:`~pincer.core.gateway.Gateway`
        The gateway for the current shard.

    Returns
    -------
    Tuple[:class:`str`, :class:`~pincer.objects.events.webhook.WebhooksUpdateEvent`]
        ``on_webhooks_update`` and a ``WebhooksUpdateEvent``
    """  # noqa: E501
    return (
        "on_webhooks_update",
        WebhooksUpdateEvent.from_dict(construct_client_dict(self, payload.data)),
    )


def export():
    return webhooks_update_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""sent when a thread is updated"""

from __future__ import annotations

from typing import TYPE_CHECKING

from ..objects import Channel
from ..utils import construct_client_dict, replace

if TYPE_CHECKING:
    from ..client import Client
    from ..core.gateway import Gateway
    from ..core.gateway import GatewayDispatch


async def thread_update_middleware(
    self: Client,
    gatewayer: Gateway,
    payload: GatewayDispatch
):
    """|coro|

    Middleware for the ``on_thread_update`` event.

    Parameters
    ----------
    payload : :class:`~pincer.core.gateway.GatewayDispatch`
        The data received from the thread update event.
    gateway : :class:`~pincer.core.gateway.Gateway`
        The gateway for the current shard.

    Returns
    -------
    Tuple[:class:`str`, :class:`~pincer.objects.guild.channel.Channel`]
        ``on_thread_update`` and an ``Channel``
    """

    channel = Channel.from_dict(construct_client_dict(self, payload.data))

    guild = self.guilds.get(channel.guild_id)

    if guild:
        guild.threads = replace(
            lambda _channel: _channel.id == channel.id,
            self.guilds[channel.guild_id].threads,
            channel,
        )
        self.channels[channel.id] = channel

    return "on_thread_update", channel


def export():
    return thread_update_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""sent when a guild is updated"""

from __future__ import annotations

from typing import TYPE_CHECKING

from ..objects import Guild
from ..utils.conversion import construct_client_dict

if TYPE_CHECKING:
    from ..client import Client
    from ..core.gateway import Gateway
    from ..core.gateway import GatewayDispatch


async def guild_update_middleware(
    self: Client,
    gateway: Gateway,
    payload: GatewayDispatch
):
    """|coro|

    Middleware for the ``on_guild_update`` event.

    Parameters
    ----------
    payload : :class:`~pincer.core.gateway.GatewayDispatch`
        The data received from the guild update event.
    gateway : :class:`~pincer.core.gateway.Gateway`
        The gateway for the current shard.

    Returns
    -------
    Tuple[:class:`str`, :class:`~pincer.objects.guild.guild.Guild`]
        ``on_guild_Update`` and an ``Guild``
    """

    guild = Guild.from_dict(construct_client_dict(self, payload.data))
    self.guilds[guild.id] = guild

    for channel in guild.channels:
        self.channels[channel.id] = channel

    return "on_guild_update", guild


def export():
    return guild_update_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from copy import copy

from ...utils.api_object import APIObject


class _Component(APIObject):
    """Represents a message component with attributes that can be modified inline"""

    def with_attrs(self, **kwargs) -> _Component:
        """
        Modifies attributes are returns an object with these attributes.

        .. code-block:: python

            some_button.with_attrs(label="A new label")

        \\*\\*kwargs
            Attributes to modify and their values.
        """
        copied_obj = copy(self)

        for key, value in kwargs.items():
            setattr(copied_obj, key, value)

        return copied_obj



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from functools import partial
from inspect import iscoroutinefunction
from typing import List

from .button import Button, ButtonStyle
from .select_menu import SelectMenu, SelectOption
from .component_handler import ComponentHandler
from ...exceptions import CommandIsNotCoroutine
from ...objects.message.emoji import Emoji
from ...utils.conversion import remove_none


def component(custom_id):
    """
    Generic handler for Message Components. Can be used with manually constructed
    :class:`~pincer.commands.components.button.Button` and
    :class:`~pincer.commands.components.select_menu.SelectMenu` objects.

    Parameters
    ---------
    custom_id : str
        The ID of the message component to handle.
    """
    def wrap(custom_id, func):
        ComponentHandler().register_id(_id=custom_id, func=func)
        return func

    return partial(wrap, custom_id)


def button(
    label: str,
    style: ButtonStyle,
    emoji: Emoji = None,
    url: str = None,
    disabled: bool = None,
    custom_id: str = None
) -> Button:
    """
    Turn a function into handler for a :class:`~pincer.commands.components.button.Button`.
    See :class:`~pincer.commands.components.button.Button` for information on parameters.

    The function will still be callable.

    .. code-block:: python

        from pincer.commands import ActionRow, Button

        class Bot(Client):

            @command
            async def send_a_button(self):
                return Message(
                    content="Click a button",
                    components=[
                        ActionRow(
                            self.button_one
                        )
                    ]
                )

            @button(label="Click me!", style=ButtonStyle.PRIMARY)
            async def button_one():
                return "Button one pressed"
    """  # noqa: E501

    def wrap(custom_id, func) -> Button:

        if not iscoroutinefunction(func):
            raise CommandIsNotCoroutine(f"`{func.__name__}` must be a coroutine.")

        if custom_id is None:
            custom_id = func.__name__

        ComponentHandler().register_id(custom_id, func)

        button = Button(
            # Hack to not override defaults in button class
            **remove_none(
                {
                    "custom_id": custom_id,
                    "style": style,
                    "label": label,
                    "disabled": disabled,
                    "emoji": emoji,
                    "url": url,
                    "_func": func
                }
            )
        )

        button.func = func
        button.__call__ = partial(func)

        return button

    return partial(wrap, custom_id)


def select_menu(
    func=None,
    options: List[SelectOption] = None,
    placeholder: str = None,
    min_values: int = None,
    max_values: int = None,
    disabled: bool = None,
    custom_id: str = None
) -> SelectMenu:
    """
    Turn a function into handler for a :class:`~pincer.commands.components.select_menu.SelectMenu`.
    See :class:`~pincer.commands.components.select_menu.SelectMenu` for information on parameters.

    The function will still be callable.

    .. code-block:: python

        from pincer.commands import button, ActionRow, ButtonStyle

        class Bot(Client):

            @command
            async def send_a_select_menu(self):
                return Message(
                    content="Choose an option",
                    components=[
                        ActionRow(
                            self.select_menu
                        )
                    ]
                )

            @select_menu(options=[
                SelectOption(label="Option 1"),
                SelectOption(label="Option 2", value="value different than label")
            ])
            async def select_menu(values: List[str]):
                return f"{values[0]} selected"

    """  # noqa: E501

    def wrap(custom_id, func) -> SelectMenu:

        if not iscoroutinefunction(func):
            raise CommandIsNotCoroutine(f"`{func.__name__}` must be a coroutine.")

        if custom_id is None:
            custom_id = func.__name__

        ComponentHandler().register_id(custom_id, func)

        menu = SelectMenu(
            # Hack to not override defaults in button class
            **remove_none(
                {
                    "custom_id": custom_id,
                    "options": options,
                    "placeholder": placeholder,
                    "min_values": min_values,
                    "max_values": max_values,
                    "disabled": disabled,
                    "_func": func
                }
            )
        )

        return menu

    if func is None:
        return partial(wrap, custom_id)

    return wrap(custom_id, func)



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from asyncio import sleep
from dataclasses import dataclass
import logging
from time import time
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Dict, Tuple
    from .http import HttpCallable

_log = logging.getLogger(__name__)


@dataclass
class Bucket:
    """Represents a rate limit bucket

    Attributes
    ----------
    limit : int
        The number of requests that can be made.
    remaining : int
        The number of remaining requests that can be made.
    reset : float
        Epoch time at which rate limit resets.
    reset_after : float
        Total time in seconds until rate limit resets.
    time_cached : float
        Time since epoch when this bucket was last saved.
    """
    limit: int
    remaining: int
    reset: float
    reset_after: float
    time_cached: float


class RateLimiter:
    """Prevents ``user`` rate limits
    Attributes
    ----------
    bucket_map : Dict[Tuple[str, :class:`~pincer.core.http.HttpCallable`], str]
        Maps endpoints and methods to a rate limit bucket
    buckets : Dict[str, :class:`~pincer.core.ratelimiter.Bucket`]
        Dictionary of buckets
    """

    def __init__(self) -> None:
        self.bucket_map: Dict[Tuple[str, HttpCallable], str] = {}
        self.buckets: Dict[str, Bucket] = {}

    def save_response_bucket(
        self,
        endpoint: str,
        method: HttpCallable,
        header: Dict
    ):
        """
        Parameters
        ----------
        endpoint : str
            The endpoint
        method : :class:`~pincer.core.http.HttpCallable`
            The method used on the endpoint (E.g. ``Get``, ``Post``, ``Patch``)
        header : :class:`aiohttp.typedefs.CIMultiDictProxy`
            The headers from the response
        """
        bucket_id = header.get("X-RateLimit-Bucket")

        if not bucket_id:
            return

        self.bucket_map[endpoint, method] = bucket_id

        self.buckets[bucket_id] = Bucket(
            limit=int(header["X-RateLimit-Limit"]),
            remaining=int(header["X-RateLimit-Remaining"]),
            reset=float(header["X-RateLimit-Reset"]),
            reset_after=float(header["X-RateLimit-Reset-After"]),
            time_cached=time()
        )

        _log.info(
            "Rate limit bucket detected: %s - %r.",
            bucket_id,
            self.buckets[bucket_id]
        )

    async def wait_until_not_ratelimited(
        self,
        endpoint: str,
        method: HttpCallable
    ):
        """|coro|
        Waits until the response no longer needs to be blocked to prevent a
        429 response because of ``user`` rate limits.

        Parameters
        ----------
        endpoint : str
            The endpoint
        method : :class:`~pincer.core.http.HttpCallable`
            The method used on the endpoint (E.g. ``Get``, ``Post``, ``Patch``)
        """
        bucket_id = self.bucket_map.get((endpoint, method))

        if not bucket_id:
            return

        bucket = self.buckets[bucket_id]
        cur_time = time()

        if bucket.remaining == 0:
            sleep_time = cur_time - bucket.time_cached + bucket.reset_after

            _log.info(
                "Waiting for %ss until rate limit for bucket %s is over.",
                sleep_time,
                bucket_id
            )

            await sleep(sleep_time)

            _log.info(
                "Message sent. Bucket %s rate limit ended.",
                bucket_id
            )



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""sent when multiple messages are deleted at once"""

from __future__ import annotations

from typing import TYPE_CHECKING

from ..objects.events.message import MessageDeleteBulkEvent
from ..utils.conversion import construct_client_dict
from ..utils.types import Coro

if TYPE_CHECKING:
    from ..client import Client
    from ..core.gateway import Gateway
    from ..core.gateway import GatewayDispatch


async def message_delete_bulk_middleware(
    self: Client,
    gateway: Gateway,
    payload: GatewayDispatch
):
    """|coro|

    Middleware for the ``on_message_delete_bulk`` event.

    Parameters
    ----------
    payload : :class:`~pincer.core.gateway.GatewayDispatch`
        The data received from the message delete bulk event
    gateway : :class:`~pincer.core.gateway.Gateway`
        The gateway for the current shard.

    Returns
    -------
    Tuple[:class:`str`, :class:`~pincer.events.message.MessageDeleteBulkEvent`]
        ``on_message_delete_bulk`` and an ``MessageDeleteBulkEvent``
    """
    return (
        "on_message_delete_bulk",
        MessageDeleteBulkEvent.from_dict(
            construct_client_dict(self, payload.data)
        ),
    )


def export() -> Coro:
    return message_delete_bulk_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from enum import Enum, auto


class GuildFeature(Enum):
    """Represents Guild Features strings.

    Attributes
    ----------
    ANIMATED_ICON : str
        Guild can have an animated icon.
    BANNER : str
        Guild can have a banner.
    COMMERCE : str
        Guild can have commerce.
    COMMUNITY : str
        Guild is a community.
    DISCOVERABLE : str
        Guild can be discovered.
    FEATURABLE : str
        Guild can be featureurable.Guild can have an invite splash.
    INVITE_SPLASH : str
        Guild can be featureurable.
    MEMBER_VERIFICATION_GATE_ENABLED : str
        Guild has member verification enabled.
    NEWS : str
        Guild has a news channel.
    PARTNERED : str
        Guild is a partner.
    PREVIEW_ENABLED : str
        Guild has preview enabled.
    VANITY_URL : str
        Guild has a vanity url.
    VERIFIED : str
        Guild is verified.
    VIP_REGIONS : str
        Guild can have VIP regions.
    WELCOME_SCREEN_ENABLED : str
        Guild has welcome screen enabled.
    TICKETED_EVENTS_ENABLED : str
        Guild has ticketed events enabled.
    MONETIZATION_ENABLED : str
        Guild has monetization enabled.
    MORE_STICKERS : str
        Guild can have more stickers.
    THREE_DAY_THREAD_ARCHIVE : str
        Guild can have three day archive time for threads.
    SEVEN_DAY_THREAD_ARCHIVE : str
        Guild can have seven day archive time for threads.
    PRIVATE_THREADS : str
        Guild can have private threads.
    """
    ANIMATED_ICON = "ANIMATED_ICON"
    BANNER = "BANNER"
    COMMERCE = "COMMERCE"
    COMMUNITY = "COMMUNITY"
    DISCOVERABLE = "DISCOVERABLE"
    FEATURABLE = "FEATURABLE"
    INVITE_SPLASH = "INVITE_SPLASH"
    MEMBER_VERIFICATION_GATE_ENABLED = "MEMBER_VERIFICATION_GATE_ENABLED"
    NEWS = "NEWS"
    PARTNERED = "PARTNERED"
    PREVIEW_ENABLED = "PREVIEW_ENABLED"
    VANITY_URL = "VANITY_URL"
    VERIFIED = "VERIFIED"
    VIP_REGIONS = "VIP_REGIONS"
    WELCOME_SCREEN_ENABLED = "WELCOME_SCREEN_ENABLED"
    TICKETED_EVENTS_ENABLED = "TICKETED_EVENTS_ENABLED"
    MONETIZATION_ENABLED = "MONETIZATION_ENABLED"
    MORE_STICKERS = "MORE_STICKERS"
    THREE_DAY_THREAD_ARCHIVE = "THREE_DAY_THREAD_ARCHIVE"
    SEVEN_DAY_THREAD_ARCHIVE = "SEVEN_DAY_THREAD_ARCHIVE"
    PRIVATE_THREADS = "PRIVATE_THREADS"
    NEW_THREAD_PERMISSIONS = "NEW_THREAD_PERMISSIONS"



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from .commands import (
    command,
    user_command,
    message_command,
    ChatCommandHandler,
    hash_app_command,
    hash_app_command_params,
)
from .arg_types import (
    CommandArg,
    Description,
    Choice,
    Choices,
    ChannelTypes,
    MaxValue,
    MinValue,
    Modifier,
)
from .components import (
    ActionRow, Button, ButtonStyle, ComponentHandler, SelectMenu, SelectOption,
    component, button, select_menu, LinkButton
)

__all__ = (
    "ActionRow", "Button", "ButtonStyle", "ChannelTypes",
    "ChatCommandHandler", "Choice", "Choices", "CommandArg",
    "ComponentHandler", "Description", "LinkButton", "MaxValue", "MinValue",
    "Modifier", "SelectMenu", "SelectOption", "button", "command", "component",
    "hash_app_command", "hash_app_command_params", "message_command",
    "select_menu", "user_command"
)



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""sent when a guild role was updated."""

from __future__ import annotations

from typing import TYPE_CHECKING

from ..objects.events.guild import GuildRoleUpdateEvent
from ..utils import Coro
from ..utils.conversion import construct_client_dict

if TYPE_CHECKING:
    from ..client import Client
    from ..core.gateway import Gateway
    from ..core.gateway import GatewayDispatch


async def guild_role_update_middleware(
    self: Client,
    gateway: Gateway,
    payload: GatewayDispatch
):
    """|coro|

    Middleware for the ``on_guild_role_update`` event.

    Parameters
    ----------
    payload : :class:`~pincer.core.gateway.GatewayDispatch`
        The data received from the guild role update event.
    gateway : :class:`~pincer.core.gateway.Gateway`
        The gateway for the current shard.

    Returns
    -------
    Tuple[:class:`str`, :class:`~pincer.objects.events.guild.GuildRoleUpdateEvent`]
        ``on_guild_role_update`` and a ``GuildRoleUpdateEvent``
    """

    event = GuildRoleUpdateEvent.from_dict(
        construct_client_dict(self, payload.data)
    )

    guild = self.guilds.get(event.guild_id)

    if guild:
        guild.roles = [
            role if role.id != event.role.id else event.role
            for role in guild.roles
        ]

    return ("on_guild_role_update", event)


def export() -> Coro:
    return guild_role_update_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""sent when an integration is deleted"""

from __future__ import annotations

from typing import TYPE_CHECKING

from ..objects.events.integration import IntegrationDeleteEvent
from ..utils.conversion import construct_client_dict
from ..utils.types import Coro

if TYPE_CHECKING:
    from ..client import Client
    from ..core.gateway import Gateway
    from ..core.gateway import GatewayDispatch


async def integration_delete_middleware(
    self: Client,
    gateway: Gateway,
    payload: GatewayDispatch
):
    """|coro|

    Middleware for the ``on_integration_delete`` event.

    Parameters
    ----------
    payload : :class:`~pincer.core.gateway.GatewayDispatch`
        The data received from the integration delete event
    gateway : :class:`~pincer.core.gateway.Gateway`
        The gateway for the current shard.

    Returns
    -------
    Tuple[:class:`str`, :class:`~pincer.events.integration.IntegrationDeleteEvent`]
        ``on_integration_delete`` and an ``IntegrationDeleteEvent``
    """  # noqa: E501
    return (
        "on_integration_delete",
        IntegrationDeleteEvent.from_dict(
            construct_client_dict(self, payload.data)
        ),
    )


def export() -> Coro:
    return integration_delete_middleware
